/* Ce programme permet de gérer une base de donnée de carte RFID

La base de donnée est le fichier texte "Bd_cartes.txt". Le fichier vide est déjà présent (il n'est pas créé
dans ce programme...il pourrait l'être). Cela permet de conserver les cartes déjà enregistrées si l'on quitte puis
si on relance l'application.

Le suivi des menus se fait dans la fenêtre texte de Processing. On pourrait envisager de passer à un affichage 
entièrement graphique.

Fonctionnalités : 
On peut visualiser la base de donnée (liste des cartes en mémoire)
On peut insérer ou supprimer des cartes soit en saisissant les numéros des cartes au clavier, soit en passant
les cartes sur le lecteur en lisant ce qui arrive sur le port COM du PC.
On peut tester si une carte est dans la base ou pas en la passant sur le lecteur.

La difficulté de gérer la saisie au clavier dans Processing a nécessité la mise en place d'une gestion de menus
(pour situer la saisie dans un contexte et réagir à la saisie clavier en fonction de ce contexte).
La fonction Void draw() est pénalisante pour l'utilisation du while() en attente de saisie clavier qui n'est plus 
scrutée dans ce cas ! Ca n'est donc pas simple à gérer. Processing n'est pas prévu pour cela, d'après ce que j'ai pu
essayer.

Le programme est livré en l'état. Il reste de petits bugs de fonctionnement (Il faut 2 détections de carte pour que 
l'afficheur graphique passe au rouge ou au vert, des enchainnements de menus sucessif) mais c'est bien peu de chose au
regard des difficultés affrontées pour atteindre l'ensemble de ces fonctionnalités.

Bonne utilisation et réutilisation du code.

Laurent Coutelier
le 13/06/2013 sur Processing2.0
laurent.coutelier@ac-nantes.fr


*/
import processing.serial.*; //pour pouvoir lire le port serie
                            // le branchement sur un port usb Série est possible avec un adaptateur

Serial lecteurcarte;    // variable "lecteur carte" de classe Serial
int menu, ancien_menu; // suivi du menu en cours
String saisie_clavier_en_cours="", saisie_clavier_validee="";
int entrer=0;//compte le nombre de fois que la touche "Entrée" est appuyée
boolean saisie_finie=false;
boolean nouvelle_touche=false;
char touche=0;
String inString;  // création d'une cahine de caractères "inString"
int compteur=0;
String code_carte_hex="", code_carte_lue=""; 
int longueur_code=22;
PFont myFont;     // création d'une police de caractères "myFont"


void setup() {
  // Taille de l'ecran 400x200 pixel
  size(400, 200); 
  //Police de caractères
  myFont = loadFont("ArialMT-18.vlw"); 
  textFont(myFont, 18);

  //configuration lecteurcarte comme port serie 0 (RS232) à 9600Bd. 
  //println(Serial.list());
  lecteurcarte = new Serial(this, Serial.list()[0], 9600);
  println("\n" +"\n");
  menu0();
}


void draw() {

  if (code_carte_lue.length()>0) {  // si une carte est passé devant le lecteur les actions dépendront du menu en cours
    switch(menu) {
      //****************************************************      
    case 0 : // Dans le menu 0, la carte est analysée (valide ou non valide) par rapport à la base de donnée 
      //test si une carte est lue
      if (code_carte_lue.length()>=longueur_code) {
        // appel la fonction booleenne "validation_carte" et test si carte valide
        if (validation_carte(code_carte_lue)) {  // Si la carte est valide, on passe la fenêtre graphique au vert pendant 3 s
          background(0, 255, 0);  //ecran vert 
          text("Carte valable", 10, 50);//imprime dans la fenêtre 400*200
          delay(3000);
          menu0(); // retour au menu0
        }
        //si carte non valide on passe la fenêtre graphique au rouge pendant 3 s
        else {
          background(255, 0, 0); //ecran rouge  
          text("Carte invalide", 10, 50);//imprime dans la fenêtre 400*200
          delay(3000);
          menu0();
        }
      }
      break;

      //*****************************************************
    case 2 :  // menu de suppression de carte 
      entrer=2;
      saisie_finie=true;
      supprime_une_carte(code_carte_lue);  // appel de la fonction de suppression de la carte dans la BD.
      saisie_clavier_en_cours="";
      saisie_clavier_validee="";
      entrer=0;
      saisie_finie=false;
      menu0();// retour au menu0
      break;

      //**************************************************
    case 3 :// dans le menu insertion d'une carte 
      entrer=2;
      saisie_finie=true;
      insere_une_carte(code_carte_lue);// appel de la fonction de Inserssion de la carte dans la BD.
      saisie_clavier_en_cours="";
      saisie_clavier_validee="";
      entrer=0;
      saisie_finie=false;
      menu0();// retour au menu0
      break;
    default :
    }
  }
  else {
    //si pas de carte  
    code_carte_lue="";
    //menu0();
    background(0); //ecran noir  
    text("insérer une carte", 10, 50);//imprime dans la fenêtre 400*200
  }

// gestion des frappe au clavier en fonction du menu en cours
  if (nouvelle_touche==true) {
    switch(menu) {
      //****************************************************      
    case 0 : // menu 0 attente d'un choix d'opération (visualisation de la BD, suppression ou ajout d'une carte)
      switch (key) {
      case 'v' :
        println();
        afficher_la_BDB();  // la visualisation de la BD est directe.
        menu0();  //retour au menu 0
        break;
      case 's' : 
        menu2();  // passage au menu 2 et affichage (suppression d'une carte)
        break;
      case 'i' : 
        menu3();   // passage au menu 3 et affichage  (ajout d'une carte)   
        break;
      default : // gestion des autres touches hors contexte
        println();
        println("Saisie non valide, veuillez recommencer");
        println();
        menu0();
      }    
      nouvelle_touche=false; // sortie du menu si la dernière touche à été traitée.
      break;

      //*****************************************************
    case 2 : //gestion des touches dans le menu de suppression
      switch (touche) {

      case 10 :  // touche Entrée
        entrer=entrer+1;  // Validation intermédiaire
        if (entrer==1) {
          println();
          println("le code de la carte saisi est " +saisie_clavier_en_cours);
          println("Appuyer sur \"Entrée\" pour confirmer ou sur \"Retour arrière\" pour annuler.");
        }      
        if (entrer==2) {  // Validation définitive -> test du numéro de carte (N° de carte non vide ) 
          saisie_clavier_validee=saisie_clavier_en_cours;
          saisie_clavier_en_cours="";
          saisie_finie=true;
          entrer=0;
          if (saisie_finie) { 
            if (saisie_clavier_validee.length()>0 && saisie_finie) {
              supprime_une_carte(saisie_clavier_validee);   // appel à la fonction de suppression du numéro de carte de la base
              saisie_clavier_en_cours="";
              saisie_clavier_validee="";
              entrer=0;
              saisie_finie=false;
            }
            else {
              println("Pas de code, pas d'enregistrement !");
              println();
            }
            menu0(); // retour au menu 0
          }
          else {
          }
        }
        break;

      case 8 :  //  gestion de l'appui sur la touche retour arrière
        println("La saisie est annulée");
        println();
        saisie_clavier_en_cours="";
        saisie_clavier_validee="";
        entrer=0;
        saisie_finie=true;
        menu0();// retour au menu 0
        break;

      default :  // sinon on gère l'appui sur les touches "normales"
        if ((touche>='A' && touche<='F' || touche>='0' && touche<='9') && saisie_clavier_en_cours.length()<=21 ) {
          saisie_clavier_en_cours=saisie_clavier_en_cours + touche; 
          print(touche);
        }
        else {  // La saisie ne permet pas d'erreur, il faut recommencer si on n'a pas tapé sur ABCDEF0123456789
          println("Votre saisie n'est pas correcte. Fin de saisie");
          println();
          saisie_clavier_en_cours="";
          saisie_clavier_validee="";
          entrer=0;
          saisie_finie=true;
          menu0();// retour au menu 0
        }
        break;
      }

      break;


      //**************************************************
    case 3 : //gestion des touches dans le menu d'inserssion d'une nouvelle carte
      switch (touche) {

      case 10 :  // touche Entrée
        entrer=entrer+1;  // Validation intermédiaire
        if (entrer==1) {
          println();
          println("le code de la carte saisi est " +saisie_clavier_en_cours);
          println("Appuyer sur \"Entrée\" pour confirmer ou sur \"Retour arrière\" pour annuler.");
        }      
        if (entrer==2) {// Validation définitive -> test du numéro de carte (N° de carte non vide ) 
          saisie_clavier_validee=saisie_clavier_en_cours;
          saisie_clavier_en_cours="";
          saisie_finie=true;
          entrer=0;
          if (saisie_finie) { 
            if (saisie_clavier_validee.length()>0 && saisie_finie) {
              insere_une_carte(saisie_clavier_validee);// appel à la fonction d'ajout du numéro de carte à la base
              saisie_clavier_en_cours="";
              saisie_clavier_validee="";
              entrer=0;
              saisie_finie=false;
            }
            else {
              println("Pas de code, pas d'enregistrement !");
              println();
            }
            menu0();// retour au menu 0
          }
          else {
          }
        }
        break;

      case 8 : //  gestion de l'appui sur la touche retour arrière
        println("La saisie est annulée");
        println();
        saisie_clavier_en_cours="";
        saisie_clavier_validee="";
        entrer=0;
        saisie_finie=true;
        menu0();// retour au menu 0
        break;


      default : // sinon on gère l'appui sur les touches "normales"
        if ((touche>='A' && touche<='F' || touche>='0' && touche<='9') && saisie_clavier_en_cours.length()<=21 ) {
          saisie_clavier_en_cours=saisie_clavier_en_cours + touche;
          print(touche);
        }
        else {// La saisie ne permet pas d'erreur, il faut recommencer si on n'a pas tapé sur ABCDEF0123456789
          println("Votre saisie n'est pas correcte. Fin de saisie");
          println();
          saisie_clavier_en_cours="";
          saisie_clavier_validee="";
          entrer=0;
          saisie_finie=true;
          menu0();// retour au menu 0
        }
      }
    }
    nouvelle_touche=false;
  }
}



//**************************************************************************************************
//**************************************************************************************************


// Fonction qui test la validitée de la carte
Boolean validation_carte(String carte_a_tester) {
  Boolean resultat_test=false;
  print("carte en cours de validation : ");//imprime sur l'écran ci-dessous

  //charge la base de données dans la variable "liste_cartes"
  String[]  liste_cartes =loadStrings("Bd_cartes.txt");

  //test si la carte est bien dans la liste des cartes valides "liste_carte"
  for (int i=0;i<liste_cartes.length;i++) {
    if (carte_a_tester.equals(liste_cartes[i])) {  // Si la carte est bonne  
      println("Carte valable"); 
      println();   
      resultat_test=true;
      code_carte_lue="";
      return resultat_test;
    }
  }
  println("non valable");// Sinon la carte n'est pas bonne
  println();  
  resultat_test=false;
  code_carte_lue="";
  return resultat_test;
}



void menu0() {
  menu=0;
  println("Que voulez faire ?");
  println("Voir la liste de carte en mémoire ? (Taper \'v\')");
  println("Supprimer une carte en mémoire ? (Taper \'s\')");
  println("Insère une carte ? (Taper \'i\')");
  println();
}

void menu1() { 
  menu=1;
}      

void menu2() {  
  menu=2;
  println("SUPPRIMER UNE CARTE");
  println("Saisir la carte au clavier (ex: ABCEDEF0123456789 ),");
  println("puis valider avec la touche \"Entrée\" annuler avec la touche \"Retour Arrière\".");
  println("Vous pouvez aussi passer la carte à saisir sur le lecteur.");
  println();
  saisie_clavier_en_cours=""; 
  saisie_clavier_validee="";
  saisie_finie=false;
}

void menu3() {  
  menu=3;
  println("AJOUTER UNE CARTE");
  println("Saisir la carte au clavier (ex: ABCEDEF0123456789 ),");
  println("puis valider avec la touche \"Entrée\" annuler avec la touche \"Retour Arrière\".");
  println("Vous pouvez aussi passer la carte à saisir sur le lecteur.");
  println();
  saisie_clavier_en_cours=""; 
  saisie_clavier_validee="";
  saisie_finie=false;
}


void afficher_la_BDB() {
  String[]  BDB_cartes =loadStrings("Bd_cartes.txt");
  println("le fichier contient " +  BDB_cartes.length + " cartes");
  println("Liste des cartes en mémoire :");
  println(BDB_cartes);
  println();
}


void supprime_une_carte(String carte_a_supprimer) {
  Boolean carte_supprimee=false;
  String[]  BDB_cartes =loadStrings("Bd_cartes.txt");  // mise en donnée tableau de String des numéros du fichier
  println("le fichier contient " +  BDB_cartes.length + " lignes");
  //  Evaluation de la présence de la carte dans la liste
  for (int i=0; i<BDB_cartes.length;i++) {
    if (BDB_cartes[i].equals(carte_a_supprimer)) {  // Si la carte est présente
      BDB_cartes[i]="";                             // On la supprime
      println("La carte est dans la base, elle va être supprimée");      
      saveStrings("Bd_cartes.txt", BDB_cartes);        // on enregistre les données du tableau modifié dans le fichier
      println("La carte a été supprimée de la base"); 
      println();     
      i=BDB_cartes.length;    // pour sortir du if() 
      carte_supprimee=true;
    }
  }
  if (!carte_supprimee) {
    println("cette carte ne figure pas dans la base");
    println();
  }
}


void insere_une_carte(String carte_a_inserer) {
  Boolean carte_inseree=false;
  Boolean carte_nouvelle=true;
  String[]  BDB_cartes =loadStrings("Bd_cartes.txt");  // mise en donnée tableau de String des numéros du fichier
  println("le fichier contient " +  BDB_cartes.length + " lignes");
  //  Evaluation de la présence de la carte dans la liste
  for (int i=0; i<BDB_cartes.length;i++) {
    if (BDB_cartes[i].equals(carte_a_inserer)) { // Si la carte est présente on fait rien
      println("La carte est déjà enregistrée");
      println();   
      carte_nouvelle=false;// La carte est déclaré comme non nouvelle
      i=BDB_cartes.length;
    }
  }
  if (carte_nouvelle) {    // Gestion de la nouvelle carte
                           // Les numéros de cartes ne sont pas classés mais les
                           // nouvelles cartes sont mises dans les trous (cartes prédemment
                           //  supprimées) ou à la fin s'il n'y a pas de trou 
    println("La carte est nouvelle, elle va être enregistrée");
    //******************************************
    //  Recherche d'une place vide dans la liste
    for (int j=0; j<BDB_cartes.length;j++) {
      if (BDB_cartes[j].equals("")) {  // recherche d'une place vide dans la liste de carte
        BDB_cartes[j]=carte_a_inserer; // mise en place du code de la carte dans le trou
        j=BDB_cartes.length;
        saveStrings("Bd_cartes.txt", BDB_cartes);  // enregistrement dans le fichier
        carte_inseree=true;
      }
    }
    //******************************************
    //  insertion à la fin de la liste
    if (carte_inseree==false) {
      //println("insertion à la fin de la liste");
      String[] carte_a_inserer2=new String [1];   // mise en tableau du n° de la carte pour pouvoir l'ajouter au tableau existant
      carte_a_inserer2[0] =carte_a_inserer;
      String[]  BDB_cartes2 = new String[BDB_cartes.length];
      BDB_cartes2 = concat(BDB_cartes, carte_a_inserer2);  // Addition de la carte (tableau de String 1 ligne) avec le tableau initial
      //println("le fichier nouveau fichier contient " +  BDB_cartes2.length + " lignes");
      saveStrings("Bd_cartes.txt", BDB_cartes2);       // on enregistre les données du tableau modifié dans le fichier
      BDB_cartes2=loadStrings("Bd_cartes.txt");        // vérification par relecture du fichier
      println("le nouveau fichier contient " +  BDB_cartes2.length + " lignes");
      carte_inseree=true;
    }
    println("La nouvelle carte est enregistrée");
    println();
  }
}




void keyPressed() {
  //println("touche clavier" +key);
  if (key>0 &&key<123) {  // Ceci filtre les appuis sur les touches comme CTRL par exemple
    nouvelle_touche=true;    //  jeton pour la gestion de la saisie en fonction des menus
    touche=key;              // récupération de la touche appuyée
    if (touche>='a' && touche<='f' ) {  // Ceci afin de remplacer les 'a' en 'A' pour ne 
      touche=(char) ((int)touche-32);   // pas avoir à appuyer sur la touche MAJ
    }
    //println("touche clavier" +touche);
  }
}

//interruption évènement port série.
void serialEvent(Serial p) {  
  //convertie la valeur lue integer (p.read)  en hexadécimale (hex(p.read) et stock dans la variable "inString" 
  inString = hex( p.read());
  //stock dans "code_carte_hex" les deux derniers caractères hexa de "inString" (entier de 4 octets)
  code_carte_hex =   code_carte_hex + inString.substring(6, 8);
  //met dans variable "code_carte_lue" les onze octets du code de la carte.
  compteur= compteur+1;
  if (compteur==longueur_code/2) { 
    code_carte_lue=code_carte_hex;
    p.clear();
    compteur=0;
    //imprime dans l'ecran ci-dessous
    println(" ");
    print("code carte lu : "); 
    print(code_carte_lue); 
    println(".");
    //vide la variable "code_carte_hex"
    code_carte_hex="";
  }
} 

